
=head1 NAME

measure5d.pm 

=head1 SYNOPSIS

This module can be used as a foundation for building 5-dimensional 
measures of association that can then be used by statistic.pl. In 
particular this module provides functions that give convenient access to  
5-d (i.e., 5gram) frequency counts as created by count.pl, as well as  
some degree of error handling that verifies the data.  

To be used in a measure module that is to be used by statistic.pl, the  
functions provided in this module must be embedded within other functions  
that adhere to the standards and naming convention described in  
Docs/NewStats.txt and is briefly summarized below.

=head1 DESCRIPTION

The functions in this module retrieve observed 5gram frequency counts and  
marginal totals, and also compute expected values. They also provide 
support for error checking of the output produced by count.pl. These 
functions are used in all the 5gram (5d) measure modules (e.g., 
ll4.pm) provided in NSP. If you are writing your own 5d measure, you can 
use these functions as well. 

With 5gram or 5d measures we use a 4x4 contingency table to store the  
frequency counts associated with each word in the 5gram, as well as the  
number of times the 5gram occurs. The notation we employ is as follows:

Marginal Frequencies:
    
 n1ppp = the number of ngrams where the first word is word1.
 np1pp = the number of ngrams where the second word is word2.
 npp1p = the number of ngrams where the third word is word3
 nppp1 = the number of ngrams where the fourth word is word4
 n2ppp = the number of ngrams where the first word is not word1. 
 np2pp = the number of ngrams where the second word is not word2.
 npp2p = the number of ngrams where the third word is not word3.
 nppp2 = the number of ngrams where the fourth words is not word4

Observed Frequencies:

 n1111 = number of times word1, word2 and word3 occur together in
         their respective positions, joint frequency.
 n1112 = number of times word1, word 2 and word3 occur in their respective
         positions but word4 does not.
 n1121 = number of times word1, word2 and word4 occur in their respective
         positions but word3 does not.
 n1122 = number of times word1 and word2 occur in their repsective positions
         but word3 and word4 do not.
 n1211 = number of times word1, word3 and word4 occur in their respective 
         positions but word2 does not.
 n1212 = number of times word1 and word3 occur in their respective positions
         but word2 and word4 do not.
 n1221 = number of times word1 and word4 occur in their respective positions
         but word2 and word3 do not
 n1222 = number of times word1 occurs in its respective position but word2, 
         word3 and word4 do not.
 n2111 = number of times word2, word3 and word4 occur in their respective 
         positions but word1 does not.
 n2112 = number of times word2 and word3 occur in their respective positions
         but word1 and word4 do not.
 n2121 = number of times word2 and word4 occur in their respective positions
         but word1 and word3 do not.
 n2122 = number of times word2 occurs in its respective position but word1, 
         word3 and word4 do not.
 n2211 = number of times word3 and word4 occur in their respective positions
         but word1 and word2 do not.
 n2212 = number of times word3 occurs in its respective position but word1, 
         word2 and word4 do not.
 n2221 = number of times word4 occurs in its respective position but word1,
         word2, and word3 do not.
 n2222 = number of times neither word1, word2, word3 or word4 occur in their
         respective positions.

Expected Frequencies:

 m1111 = expected number of times word1, word2 and word3 occur together in
         their respective positions, joint frequency.
 m1112 = expected number of times word1, word 2 and word3 occur in their respective
         positions but word4 does not.
 m1121 = expected number of times word1, word2 and word4 occur in their respective
         positions but word3 does not.
 m1122 = expected number of times word1 and word2 occur in their repsective positions
         but word3 and word4 do not.
 m1211 = expected number of times word1, word3 and word4 occur in their respective 
         positions but word2 does not.
 m1212 = expected number of times word1 and word3 occur in their respective positions
         but word2 and word4 do not.
 m1221 = expected number of times word1 and word4 occur in their respective positions
         but word2 and word3 do not
 m1222 = expected number of times word1 occurs in its respective position but word2, 
         word3 and word4 do not.
 m2111 = expected number of times word2, word3 and word4 occur in their respective 
         positions but word1 does not.
 m2112 = expected number of times word2 and word3 occur in their respective positions
         but word1 and word4 do not.
 m2121 = expected number of times word2 and word4 occur in their respective positions
         but word1 and word3 do not.
 m2122 = expected number of times word2 occurs in its respective position but word1, 
         word3 and word4 do not.
 m2211 = expected number of times word3 and word4 occur in their respective positions
         but word1 and word2 do not.
 m2212 = expected number of times word3 occurs in its respective position but word1, 
         word2 and word4 do not.
 m2221 = expected number of times word4 occurs in its respective position but word1,
         word2, and word3 do not.
 m2222 = expected number of times neither word1, word2, word3 or word4 occur in their
         respective positions.
   
=head2 Functions Included in measure5d:

=over 4

=item 1 

The intializeStatistic function performs initialization before 
the actual computation of measures of association begin. It also 
verifies that the input consists of 5grams, and that it has all 
the frequency combinations needed to do the computations. The
frequency combinations that are required for this module are 
n1111,n1ppp,np1pp,npp1p,nppp1,nn11pp,n1p1p,n1pp1,np11p,np1p1,
npp11,n111p,n11p1,n1p11,np111.

initializeStatistic() is passed the following input parameters:

     1) The ngram size. For 5d (5gram) measures this will be 5. 
     2) The total number of 5grams in the corpus (npppp).
     3) The number of frequency combinations. 
     4) A 5-d array containing the frequency combinations.

Each row  of the array in 4) represents a single frequency combination.  
On a given row, the first element denotes the number of indices on this  
row, say 'n'. This is followed by the 'n' values that correspond to the 
indices included in the frequency combination. (For more details on   
frequency combinations, see README.pod). To use this module, the joint
frequency, n11,  as well as the marginal frequencies, n1p and np1, are
required in order to calculate the expected values.

This function does not return any values. If an error occurs, it can be 
detected by statistic.pl using the errorCode and errorString functions 
described below.   

=item 2

The getObservedValues function takes as input an array containing the 
frequency values for a 5gram as found by count.pl. This will include: 
n1111,n1ppp,np1pp,npp1p,nppp1,nn11pp,n1p1p,n1pp1,np11p,np1p1,npp11,n111p,
n11p1,n1p11,np111. The size of this array is guaranteed to be exactly the 
same as the third parameter passed to the function initializeStatistic() 
function above. 

The getObservedValues function verifies that the marginal frequencies 
n1ppp,np1pp,npp1p,nppp1,nn11pp,n1p1p,n1pp1,np11p,np1p1,npp11,n111p,
n11p1,n1p11, and np111 are consistent with the value of the joint 
frequency of the 5gram, n1111. If they are not consistent, it sets an 
error code and error message, and the function returns.

If the marginals are valid, it computes the observed values for the   
remaining cells in the 5d table ( n1112, n1121, n1122, n1211, n1212,
n1221, n1222, n2111, n2112, n2121, n2122, n2211, n2212, n2221, n2222 )
based on these marginal totals and the joint frequency and returns an 
array containing these total in the order above.

=item 3

The function calculateExpectedValues() calculates the expected 
values of the cells in the contingency table based on the marginal 
frequencies and the total sample size. The expected values are estimated 
based on the assumption that the two words in the 5gram are independent.

The function returns these values in an array ordered as follows:  
m1112, m1121, m1122, m1211, m1212, m1221, m1222, m2111, m2112, m2121, 
m2122, m2211, m2212, m2221, m2222 

=item 4

The function getMarginalValues will return the marginal frequencies in the
order of: n1ppp,np1pp,npp1p,nppp1,nn11pp,n1p1p,n1pp1,np11p,np1p1,npp11,n111p,
n11p1,n1p11, and np111.

=item 5

The function getTotal5grams() returns the total number of 5grams in the 
corpus (npppp).

=item 6

The function errorCode() returns 0 if the last operation was successful.
It will return an integer starting with 1 if the last operation failed. 
This indicates that statistic.pl should abort. It will return an integer 
starting with 2 to indicate a warning should be issued. This does not 
cause cause statistic.pl to abort. However, a warning after 
calculateStatistic() will cause the 5gram which generated that warning to 
be ignored by statistic.pl.

=item 7

The function errorString() returns the text of an error message. 

=back

=head2 Writing your own statistic module

Any measure module being used by statistic.pl must follow this  
convention. In order to make it easier to build 5d measures, we provide 
a 5d specific functions that can be embedded within the measure module in 
order to carry out common operations in calculating the values of such 
measures. 

=over 4

=item 1 

The filename should have an extension of .pm. Usually the name of 
the file should be Statistic.pm, where "Statistic" is the name of the
particular statistic you are writing.

=item 2 

Let us say you have named your file Statistic.pm. The first line of the
file should declare that its a package of the same name as the filename.
Thus the first line of the file Statistic.pm should be...

   package Statistic;

=item 3

To use the measure5d.pm module you need to include it in your 
Statistic.pm module. 

A small code snippet to ensure that it is included is as follows:

 #  Check to see if the statistic.pm module can see the
 #  measure5d.pm module ... if not see if it can be found.

 my $module = "measure5d.pm"; my $modulename = "measure5d.pm";
 if( !( -f $modulename ) ) {
    my $found = 0;
    #  Check each of the PATHS to see if the module is there
    foreach (split(/:/, $ENV{PATH})) {
 	$module = $_ . "/" . $modulename;
	if ( -f $module ) { $found = 1; last; }
    }
    # if still not found anywhere, quit!
    if ( ! $found ) { print "Could not find $modulename.\n"; exit; }
 }

 # IMPORTANT : now include the module into the current package    
 require $module;

=item 4 

You need to implement at least two functions in your package

   i)   initializeStatistic()
   ii)  calculateStatistic()


Function initializeStatistic() is passed the a set of parameters 
that include the 5gram size which is 5, the total number of 
5grams in the corpus, the number of frequency combinations 
and an array containing the frequency combinations. More detail 
of these parameters are described above in the description of 
the measure5d::initializeStatistic function.

These paramters can be passed directly into the measure5d.pm
module's function measure5d::intializeStatistic. For example: 

 sub initializeStatistic {
 
     measure5d::initializeStatistic(@_);
 
 }

where @_ contains the input parameters.

This function is called before any calls to the function
calculateStatistic() and can be used by the statistic library 
to set up any values that may be required for the calculations
later. This function is not expected to return anything. If an 
error occurs, it can be reported through the mechanisms described 
below. 

The other mandatory function is calculateStatistic(). This is
passed an array containing the frequency values for an ngram as
found in the input n-gram file. 

Function calculateStatistic() is expected to return a (possibly
floating) value as the value of the statistical measure calculated
using the frequency values passed to it. 

There exists two main functions in the module measure5d.pm in order
to help calculate the 5gram statistic. 

   1.  measure5d::getObservedValues(@frequencies)
   2.  measure5d::getExpectedValues();

The function measure5d::getObservedValues will return the list of
observed values from the given 5gram. If it does not then there
existed an error in the calculation of these values and zero should
be returned. An example of how this can be used is as follows:
  
if( !( ($n1111, $n1112, $n1121, $n1121, $n1122, $n1211, $n1212, 
$n1221, $n1222, $n2111, $n2112, $n2121, $n2122, $n2211, $n2212, 
$n2221, $n2222) = measure5d::getObservedValues(@_) ) ) {
	return(0);
    }
 

where @_ is the parameters sent to calculateStatistic from statistic.pl.
A more detailed description of this function can be seen above.

The function measure5d::getExpectedValues will return the list of
expected values from the given 5gram. If it does not then there 
existed an error in the calculation of these values and zero should
be returned. An example of how this can be used is as follows:

if( !( ($m1111, $m1112, $m1121, $m1121, $m1122, $m1211, $m1212, 
$m1221, $m1222, $m2111, $m2112, $m2121, $m2122, $m2211, $m2212, 
$m2221, $m2222) = measure5d::getExpectedValues() ) ) {
	return(0);
    }

When a library is loaded, statistic.pl checks for initializeStatistc
and calcualteStatistic functions: if they are not implemented, then 
an error is reported and the program quits. 

=item 5 

Program statistic.pl also supports three other functions that are
not mandatory, but may be implemented by the user. These are:

     i) errorCode()
    ii) errorString()
   iii) getStatisticName()

Function errorCode, if implemented, is called immediately after 
the call to function initializeStatistic() and immediately after 
every call to function calculateStatistic(). 

The measure5d.pm module implements both measure5d::errorCode() 
and measure5d::errorString().

The errorCode() and errorString() methods that are implmented in
your Statistic.pm modoule can return the value returned by the 
measure5d::errorCode() and measure5d::errorString() functions.

An example of this is below:

sub errorCode
{ 
  return measure5d::errorCode();
}

sub errorString
{
    return measure5d::errorString();
}

The third function that may be implemented is getStatisticName(). 
If this function is implemented, it is expected to return a string 
containing the name of the statistic being implmented. This string 
is used in the formatted output of statistic.pl. If this function 
is not implemented, then the statistic file name entered on the 
commandline is used in the formatted output. 

Note that all three functions described in this section are first
checked for existence before being called. So, if the user elects
to not implement these functions, no harm will be done. However, we
strongly recommend the implementation of at least the function
errorCode() since this is the only way for the statistic library to
report errors to the user. 

=item 6 

Having implemented the two mandatory functions (in point 3 above)
and zero or more of the three non-mandatory functions (in point 4
above), one must make these functions available outside the
package. To do so, one has to export them, thusly. 

For this, first include the Exporter package by including the
following line in the program

   require Exporter;

Now include the following line to inherit Exporter's functions:

   @ISA = qw ( Exporter );

Now export the various functions implemented so that they are
accessible outside this package, by adding the following line
(assume that you have implemented only the two mandatory
functions): 

   @EXPORT = qw( initializeStatistic calculateStatistic );

If you implement say the errorCode() and errorString() functions
too, you may export them like so: 

   @EXPORT = qw( initializeStatistic calculateStatistic errorCode errorString );
   
Note that the user may implement other functions too, and may
export them if he so wishes, but since statistic.pl is not
expecting anything besides the five functions above, doing so would
have no effect on statistic.pl. 

=item 7 

Finally, at the end of everything, add the line

   1;

   This will ensure that the LAST line of the file returns a 
   true value, and is necessary so that when this package is 
   loaded, it returns a TRUE value.

=back

=head2 Errors to look out for:

=over 4

=item 1 

The filename does not end with a .pm.

=item 2 

The rest of the filename (besides the extension) does not match the package
name (declared in the first line of the file). Remember its case sensitive!

=item 3 

The five functions (2 mandatory, 3 non-mandatory) must have their
names match EXACTLY with those shown above. Again, names are all
case sensitive.

=item 4 

The last line of the file is not "1;". This is necessary, and easily
overlooked!

=back

=head1 AUTHORS

    Ted Pedersen (tpederse@umn.edu)
    Satanjeev Banerjee <banerjee@cs.cmu.edu>
    Bridget McInnes (bthomson@d.umn.edu)
    
=head1 BUGS

=head1 SEE ALSO

 home page:    http://www.d.umn.edu/~tpederse/nsp.html

 mailing list: http://groups.yahoo.com/group/ngram/

=head1 COPYRIGHT

Copyright (C) 2004 Ted Pedersen and Bridget McInnes

Permission is granted to copy, distribute and/or modify this  document  
under the terms of the GNU Free Documentation License, Version 1.2 or  any  
later version published by the Free Software Foundation; with no  
Invariant Sections, no Front-Cover Texts, and no Back-Cover Texts.

Note: a copy of the GNU Free Documentation License is available on the web   
at L<http://www.gnu.org/copyleft/fdl.html> and is included in this    
distribution as FDL.txt. 

=cut

package measure5d;
require Exporter;

@ISA = qw ( Exporter );
@EXPORT = qw( initializeStatistic getObservedValues getExpectedValues getMarginalTotals getTotal5grams errorCode errorString );



# function to set up various variables before the actual computation
# starts. also to check if we are being given 5grams, and if our
# frequency combinations are enough to do the computation

sub initializeStatistic
{
    ($ngram, $nppppp, $combIndex, @freqComb) = @_;
    
    $errorCodeNumber = 0;
    $errorMessage = "";
    
    # check if ngram !=5 . statistic only defined for ngram = 5. 
    if ($ngram != 5) {
	$errorMessage = "This statistic is only available for 5grams!";
	$errorCodeNumber = 1;  return;
    }
    
     # The total number of 5grams should not be less than equal to 0
    if ($nppppp <= 0) { 
	$errorMessage = "Total number of 5grams ($nppppp) must be greater than 0.";
	$errorCodeNumber = 10;  return;
    }

    # figure out from the @freqComb array if the frequency values we
    # are going to get are indeed the ones we need. 
    # while we figure this out, we shall also note which of the indices of the 
    # array passed to function calculateStatistic are the ones we want.

    for (my $i = 0; $i < $combIndex; $i++)
    {
	$str="";
        foreach(@{$freqComb[$i]}[1..$freqComb[$i][0]]) { $str.=$_."#"; }
        if($str eq "0#1#2#3#4#"){  $n11111Index=$i; }
        if($str eq "0#")        {  $n1ppppIndex=$i; }
        if($str eq "1#")        {  $np1pppIndex=$i; }
	if($str eq "2#")        {  $npp1ppIndex=$i; }
	if($str eq "3#")        {  $nppp1pIndex=$i; }
	if($str eq "4#")        {  $npppp1Index=$i; }
        if($str eq "0#1#")      {  $n11pppIndex=$i; }
	if($str eq "0#2#")      {  $n1p1ppIndex=$i; }
	if($str eq "0#3#")      {  $n1pp1pIndex=$i; }
	if($str eq "0#4#")      {  $n1ppp1Index=$i; }
        if($str eq "1#2#")      {  $np11ppIndex=$i; }
        if($str eq "1#3#")      {  $np1p1pIndex=$i; }
	if($str eq "1#4#")      {  $np1pp1Index=$i; }
        if($str eq "2#3#")      {  $npp11pIndex=$i; }
	if($str eq "2#4#")      {  $npp1p1Index=$i; }
	if($str eq "3#4#")      {  $nppp11Index=$i; }
        if($str eq "0#1#2#")    {  $n111ppIndex=$i; }
        if($str eq "0#1#3#")    {  $n11p1pIndex=$i; }
	if($str eq "0#1#4#")    {  $n11pp1Index=$i; }
        if($str eq "0#2#3#")    {  $n1p11pIndex=$i; }
	if($str eq "0#2#4#")    {  $n1p1p1Index=$i; }
	if($str eq "0#3#4#")    {  $n1pp11Index=$i; }
        if($str eq "1#2#3#")    {  $np111pIndex=$i; }
	if($str eq "1#2#4#")    {  $np11p1Index=$i; }
	if($str eq "1#3#4#")    {  $np1p11Index=$i; }
	if($str eq "2#3#4#")    {  $npp111Index=$i; }
	if($str eq "0#1#2#3#")  {  $n1111pIndex=$i; }
	if($str eq "0#1#2#4#")  {  $n111p1Index=$i; }
	if($str eq "0#1#3#4#")  {  $n11p11Index=$i; }
	if($str eq "0#2#3#4#")  {  $n1p111Index=$i; }
	if($str eq "1#2#3#4#")  {  $np1111Index=$i; }

    }
    
    if (!(defined $n11111Index)) { $errorCodeNumber = 100; $errorMessage = "Frequency combination \"0 1 2 3 4\" missing!\n"; }
    if (!(defined $n1ppppIndex)) { $errorCodeNumber = 101; $errorMessage = "Frequency combination \"0\" missing!\n";   }
    if (!(defined $np1pppIndex)) { $errorCodeNumber = 102; $errorMessage = "Frequency combination \"1\" missing!\n";   }
    if (!(defined $npp1ppIndex)) { $errorCodeNumber = 103; $errorMessage = "Frequency combination \"2\" missing!\n";   }
    if (!(defined $nppp1pIndex)) { $errorCodeNumber = 104; $errorMessage = "Frequency combination \"3\" missing!\n";   }
    if (!(defined $npppp1Index)) { $errorCodeNumber = 104; $errorMessage = "Frequency combination \"4\" missing!\n";   }
    if (!(defined $n11pppIndex)) { $errorCodeNumber = 105; $errorMessage = "Frequency combination \"0 1\" missing!\n"; }
    if (!(defined $n1p1ppIndex)) { $errorCodeNumber = 106; $errorMessage = "Frequency combination \"0 2\" missing!\n"; }
    if (!(defined $n1pp1pIndex)) { $errorCodeNumber = 107; $errorMessage = "Frequency combination \"0 3\" missing!\n"; }
    if (!(defined $n1ppp1Index)) { $errorCodeNumber = 108; $errorMessage = "Frequency combination \"0 4\" missing!\n"; }
    if (!(defined $np11ppIndex)) { $errorCodeNumber = 109; $errorMessage = "Frequency combination \"1 2\" missing!\n"; }
    if (!(defined $np1p1pIndex)) { $errorCodeNumber = 110; $errorMessage = "Frequency combination \"1 3\" missing!\n"; }
    if (!(defined $np1pp1Index)) { $errorCodeNumber = 111; $errorMessage = "Frequency combination \"1 4\" missing!\n"; }
    if (!(defined $npp11pIndex)) { $errorCodeNumber = 112; $errorMessage = "Frequency combination \"2 3\" missing!\n"; }
    if (!(defined $npp1p1Index)) { $errorCodeNumber = 113; $errorMessage = "Frequency combination \"2 4\" missing!\n"; }
    if (!(defined $nppp11Index)) { $errorCodeNumber = 114; $errorMessage = "Frequency combination \"3 4\" missing!\n"; }
    if (!(defined $n111ppIndex)) { $errorCodeNumber = 114; $errorMessage = "Frequency combination \"0 1 2\" missing!\n"; }
    if (!(defined $n11p1pIndex)) { $errorCodeNumber = 114; $errorMessage = "Frequency combination \"0 1 3\" missing!\n"; }
    if (!(defined $n11pp1Index)) { $errorCodeNumber = 114; $errorMessage = "Frequency combination \"0 1 4\" missing!\n"; }
    if (!(defined $n1p11pIndex)) { $errorCodeNumber = 114; $errorMessage = "Frequency combination \"0 2 3\" missing!\n"; }
    if (!(defined $n1p1p1Index)) { $errorCodeNumber = 114; $errorMessage = "Frequency combination \"0 2 4\" missing!\n"; }
    if (!(defined $n1pp11Index)) { $errorCodeNumber = 114; $errorMessage = "Frequency combination \"0 3 4\" missing!\n"; }
    if (!(defined $np111pIndex)) { $errorCodeNumber = 114; $errorMessage = "Frequency combination \"1 2 3\" missing!\n"; }
    if (!(defined $np11p1Index)) { $errorCodeNumber = 114; $errorMessage = "Frequency combination \"1 2 4\" missing!\n"; }
    if (!(defined $np1p11Index)) { $errorCodeNumber = 114; $errorMessage = "Frequency combination \"1 3 4\" missing!\n"; }
    if (!(defined $npp111Index)) { $errorCodeNumber = 114; $errorMessage = "Frequency combination \"2 3 4\" missing!\n"; }
    if (!(defined $n1111pIndex)) { $errorCodeNumber = 114; $errorMessage = "Frequency combination \"0 1 2 3\" missing!\n"; }
    if (!(defined $n111p1Index)) { $errorCodeNumber = 114; $errorMessage = "Frequency combination \"0 1 2 4\" missing!\n"; }
    if (!(defined $n11p11Index)) { $errorCodeNumber = 114; $errorMessage = "Frequency combination \"0 1 3 4\" missing!\n"; }
    if (!(defined $n1p111Index)) { $errorCodeNumber = 114; $errorMessage = "Frequency combination \"0 2 3 4\" missing!\n"; }
    if (!(defined $np1111Index)) { $errorCodeNumber = 114; $errorMessage = "Frequency combination \"1 2 3 4\" missing!\n"; }

}

# function to ensure the marginal values are reasonable with
# respect to each other and if so calculate the observed values
sub getObservedValues() 
{

    $n11111=$_[$n11111Index];    $n1pppp=$_[$n1ppppIndex]; 
    $np1ppp=$_[$np1pppIndex];    $npp1pp=$_[$npp1ppIndex]; 
    $nppp1p=$_[$nppp1pIndex];    $npppp1=$_[$npppp1Index]; 
    $n11ppp=$_[$n11pppIndex];    $n1p1pp=$_[$n1p1ppIndex]; 
    $n1pp1p=$_[$n1pp1pIndex];    $n1ppp1=$_[$n1ppp1Index]; 
    $np11pp=$_[$np11ppIndex];    $np1p1p=$_[$np1p1pIndex]; 
    $np1pp1=$_[$np1pp1Index];    $npp11p=$_[$npp11pIndex]; 
    $npp1p1=$_[$npp1p1Index];    $nppp11=$_[$nppp11Index]; 
    $n111pp=$_[$n111ppIndex];    $n11p1p=$_[$n11p1pIndex]; 
    $n11pp1=$_[$n11pp1Index];    $n1p11p=$_[$n1p11pIndex]; 
    $n1p1p1=$_[$n1p1p1Index];    $n1pp11=$_[$n1pp11Index]; 
    $np111p=$_[$np111pIndex];    $np11p1=$_[$np11p1Index]; 
    $np1p11=$_[$np1p11Index];    $npp111=$_[$npp111Index]; 
    $n1111p=$_[$n1111pIndex];    $n111p1=$_[$n111p1Index]; 
    $n11p11=$_[$n11p11Index];    $n1p111=$_[$n1p111Index]; 
    $np1111=$_[$np1111Index];
    
    $n11112=$n1111p-$n11111;
    $n11121=$n111p1-$n11111;
    $n11122=$n111pp-$n11111-$n11121-$n11112;
    $n11211=$n11p11-$n11111;
    $n11212=$n11p1p-$n11111-$n11211-$n11112;
    $n11221=$n11pp1-$n11111-$n11121-$n11211;
    $n11222=$n11ppp-$n11111-$n11211-$n11121-$n11112;
    $n12111=$n1p111-$n11111;
    $n12112=$n1p11p-$n11111-$n12111-$n11112;
    $n12121=$n1p1p1-$n11111-$n12111-$n11121;
    $n12122=$n1p1pp-$n11111-$n12111-$n11121-$n11112;
    $n12211=$n1pp11-$n11111-$n12111-$n11211;
    $n12212=$n1pp1p-$n11111-$n12111-$n11211-$n11112;
    $n12221=$n1ppp1-$n11111-$n12111-$n11211-$n11121;
    $n12222=$n1pppp-$n11111-$n12111-$n11211-$n11121-$n11112;
    $n21111=$np1111-$n11111;
    $n21112=$np111p-$n11111-$n21111-$n11112;
    $n21121=$np11p1-$n11111-$n21111-$n11121;
    $n21122=$np11pp-$n11111-$n21111-$n11121-$n11112;
    $n21211=$np1p11-$n11111-$n21111-$n11211;
    $n21212=$np1p1p-$n11111-$n21111-$n11211-$n11112;
    $n21221=$np1pp1-$n11111-$n21111-$n11211-$n11121;
    $n21222=$np1ppp-$n11111-$n21111-$n11211-$n11121-$n11112;
    $n22111=$npp111-$n11111-$n21111-$n12111;
    $n22112=$npp11p-$n11111-$n21111-$n12111-$n11112;
    $n22121=$npp1p1-$n11111-$n21111-$n12111-$n11121;
    $n22122=$npp1pp-$n11111-$n21111-$n12111-$n11121-$n11112;
    $n22211=$nppp11-$n11111-$n21111-$n12111-$n11211;
    $n22212=$nppp1p-$n11111-$n21111-$n12111-$n11211-$n11112;
    $n22221=$npppp1-$n11111-$n21111-$n12111-$n11211-$n11121;
    $n22222=$nppppp-($n11111+$n11112+$n11121+$n11122+$n11211+$n11212+$n11221+
		     $n11222+$n12111+$n12112+$n12121+$n12122+$n12211+$n12212+
		     $n12221+$n12222+$n21111+$n21112+$n21121+$n21122+$n21211+
		     $n21212+$n21221+$n21222+$n22111+$n22112+$n22121+$n22122+
		     $n22211+$n22212+$n22221);


    $n2pppp=$nppppp-$n1pppp; $np2ppp=$nppppp-$np1ppp;  $npp2pp=$nppppp-$npp1pp; 
    $nppp2p=$nppppp-$nppp1p; $npppp2=$nppppp-$npppp1;
     
    # n11111 should be greater than equal to zero 
    if ($n11111< 0)    {
	$errorMessage = "Frequency value ($n11111) must not be negative.";
	$errorCodeNumber = 200;  return;
    }
    
    # n11111 frequency should be less than or equal to total5grams
    if ($n11111> $nppppp) {
	$errorMessage = "Frequency value ($n11111) must not exceed total number of 5grams.";
	$errorCodeNumber = 201;  return;
    }
    
    # n11112 should be greater than equal to zero 
    if ($n11112< 0)    {
	$errorMessage = "Frequency value ($n11112) must not be negative.";
	$errorCodeNumber = 202;  return;
    }
    
    # n11112 frequency should be less than or equal to total5grams
    if ($n11121> $nppppp) {
	$errorMessage = "Frequency value ($n11112) must not exceed total number of 5grams.";
	$errorCodeNumber = 203;  return;
    }
    
    # n11121 should be greater than equal to zero 
    if ($n11121< 0)    {
	$errorMessage = "Frequency value ($n11121) must not be negative.";
	$errorCodeNumber = 204;  return;
    }
    
    # n11121 frequency should be less than or equal to total5grams
    if ($n11121> $nppppp) {
	$errorMessage = "Frequency value ($n11121) must not exceed total number of 5grams.";
	$errorCodeNumber = 205;  return;
    }
    
    # n11122 should be greater than equal to zero 
    if ($n11122< 0)    {
	$errorMessage = "Frequency value ($n11122) must not be negative.";
	$errorCodeNumber = 206;  return;
    }
    
    # n11122 frequency should be less than or equal to total5grams
    if ($n11122> $nppppp) {
	$errorMessage = "Frequency value ($n11122) must not exceed total number of 5grams.";
	$errorCodeNumber = 207;  return;
    }

    # n11111 should be greater than equal to zero 
    if ($n11211< 0)    {
	$errorMessage = "Frequency value ($n11211) must not be negative.";
	$errorCodeNumber = 208;  return;
    }
    
    # n11211 frequency should be less than or equal to total5grams
    if ($n11211> $nppppp) {
	$errorMessage = "Frequency value ($n11211) must not exceed total number of 5grams.";
	$errorCodeNumber = 209;  return;
    }

    # n11212 should be greater than equal to zero 
    if ($n11212< 0)    {
	$errorMessage = "Frequency value ($n11212) must not be negative.";
	$errorCodeNumber = 210;  return;
    }
    
    # n11212 frequency should be less than or equal to total5grams
    if ($n11212> $nppppp) {
	$errorMessage = "Frequency value ($n11212) must not exceed total number of 5grams.";
	$errorCodeNumber = 211;  return;
    }

    # n11221 should be greater than equal to zero 
    if ($n11221< 0)    {
	$errorMessage = "Frequency value ($n11221) must not be negative.";
	$errorCodeNumber = 212;  return;
    }
    
    # n11221 frequency should be less than or equal to total5grams
    if ($n11221> $nppppp) {
	$errorMessage = "Frequency value ($n11221) must not exceed total number of 5grams.";
	$errorCodeNumber = 213;  return;
    }

    # n11222 should be greater than equal to zero 
    if ($n11222< 0)    {
	$errorMessage = "Frequency value ($n11222) must not be negative.";
	$errorCodeNumber = 214;  return;
    }
    
    # n11222 frequency should be less than or equal to total5grams
    if ($n11222> $nppppp) {
	$errorMessage = "Frequency value ($n11222) must not exceed total number of 5grams.";
	$errorCodeNumber = 215;  return;
    }

    # n12111 should be greater than equal to zero 
    if ($n12111< 0)    {
	$errorMessage = "Frequency value ($n12111) must not be negative.";
	$errorCodeNumber = 216;  return;
    }
    
    # n12111 frequency should be less than or equal to total5grams
    if ($n12111> $nppppp) {
	$errorMessage = "Frequency value ($n12111) must not exceed total number of 5grams.";
	$errorCodeNumber = 217;  return;
    }
    
    # n12112 should be greater than equal to zero 
    if ($n12112< 0)    {
	$errorMessage = "Frequency value ($n12112) must not be negative.";
	$errorCodeNumber = 218;  return;
    }
    
    # n12112 frequency should be less than or equal to total5grams
    if ($n12112> $nppppp) {
	$errorMessage = "Frequency value ($n12112) must not exceed total number of 5grams.";
	$errorCodeNumber = 219;  return;
    }

    # n12121 should be greater than equal to zero 
    if ($n12121< 0)    {
	$errorMessage = "Frequency value ($n12121) must not be negative.";
	$errorCodeNumber = 220;  return;
    }
    
    # n12121 frequency should be less than or equal to total5grams
    if ($n12121> $nppppp) {
	$errorMessage = "Frequency value ($n12121) must not exceed total number of 5grams.";
	$errorCodeNumber = 221;  return;
    }
    
    # n12122 should be greater than equal to zero 
    if ($n12122< 0)    {
	$errorMessage = "Frequency value ($n12122) must not be negative.";
	$errorCodeNumber = 222;  return;
    }
    
    # n12122 frequency should be less than or equal to total5grams
    if ($n12122> $nppppp) {
	$errorMessage = "Frequency value ($n12122) must not exceed total number of 5grams.";
	$errorCodeNumber = 223;  return;
    }

    # n12211 should be greater than equal to zero 
    if ($n12211< 0)    {
	$errorMessage = "Frequency value ($n12211) must not be negative.";
	$errorCodeNumber = 224;  return;
    }
    
    # n12211 frequency should be less than or equal to total5grams
    if ($n12211> $nppppp) {
	$errorMessage = "Frequency value ($n12211) must not exceed total number of 5grams.";
	$errorCodeNumber = 225;  return;
    }

    # n12212 should be greater than equal to zero 
    if ($n12212< 0)    {
	$errorMessage = "Frequency value ($n12212) must not be negative.";
	$errorCodeNumber = 226;  return;
    }
    
    # n12212 frequency should be less than or equal to total5grams
    if ($n12212> $nppppp) {
	$errorMessage = "Frequency value ($n12212) must not exceed total number of 5grams.";
	$errorCodeNumber = 227;  return;
    }
    
    # n12221 should be greater than equal to zero 
    if ($n12221< 0)    {
	$errorMessage = "Frequency value ($n12221) must not be negative.";
	$errorCodeNumber = 228;  return;
    }
    
    # n12221 frequency should be less than or equal to total5grams
    if ($n12221> $nppppp) {
	$errorMessage = "Frequency value ($n12221) must not exceed total number of 5grams.";
	$errorCodeNumber = 229;  return;
    }

    # n12222 should be greater than equal to zero 
    if ($n12222< 0)    {
	$errorMessage = "Frequency value ($n12222) must not be negative.";
	$errorCodeNumber = 230;  return;
    }
    
    # n12222 frequency should be less than or equal to total5grams
    if ($n12222> $nppppp) {
	$errorMessage = "Frequency value ($n12222) must not exceed total number of 5grams.";
	$errorCodeNumber = 231;  return;
    }
    
    # n21111 should be greater than equal to zero 
    if ($n21111< 0)    {
	$errorMessage = "Frequency value ($n21111) must not be negative.";
	$errorCodeNumber = 232;  return;
    }
    
    # n21111 frequency should be less than or equal to total5grams
    if ($n21111> $nppppp) {
	$errorMessage = "Frequency value ($n21111) must not exceed total number of 5grams.";
	$errorCodeNumber = 232;  return;
    }
    
    # n21112 should be greater than equal to zero 
    if ($n21112< 0)    {
	$errorMessage = "Frequency value ($n21112) must not be negative.";
	$errorCodeNumber = 233;  return;
    }
    
    # n21112 frequency should be less than or equal to total5grams
    if ($n21112> $nppppp) {
	$errorMessage = "Frequency value ($n21112) must not exceed total number of 5grams.";
	$errorCodeNumber = 234;  return;
    }

    # n21121 should be greater than equal to zero 
    if ($n11111< 0)    {
	$errorMessage = "Frequency value ($n11111) must not be negative.";
	$errorCodeNumber = 235;  return;
    }
    
    # n21121 frequency should be less than or equal to total5grams
    if ($n21121> $nppppp) {
	$errorMessage = "Frequency value ($n21121) must not exceed total number of 5grams.";
	$errorCodeNumber = 236;  return;
    }

    # n21122 should be greater than equal to zero 
    if ($n21122< 0)    {
	$errorMessage = "Frequency value ($n21122) must not be negative.";
	$errorCodeNumber = 237;  return;
    }
    
    # n21122 frequency should be less than or equal to total5grams
    if ($n21122> $nppppp) {
	$errorMessage = "Frequency value ($n21122) must not exceed total number of 5grams.";
	$errorCodeNumber = 238;  return;
    }
    
    # n21211 should be greater than equal to zero 
    if ($n21211< 0)    {
	$errorMessage = "Frequency value ($n21211) must not be negative.";
	$errorCodeNumber = 239;  return;
    }
    
    # n21211 frequency should be less than or equal to total5grams
    if ($n21211> $nppppp) {
	$errorMessage = "Frequency value ($n21211) must not exceed total number of 5grams.";
	$errorCodeNumber = 240;  return;
    }
    
    # n21212 should be greater than equal to zero 
    if ($n21212< 0)    {
	$errorMessage = "Frequency value ($n11111) must not be negative.";
	$errorCodeNumber = 241;  return;
    }
    
    # n21212 frequency should be less than or equal to total5grams
    if ($n21212> $nppppp) {
	$errorMessage = "Frequency value ($n21212) must not exceed total number of 5grams.";
	$errorCodeNumber = 242;  return;
    }

    # n21221 should be greater than equal to zero 
    if ($n21221< 0)    {
	$errorMessage = "Frequency value ($n21221) must not be negative.";
	$errorCodeNumber = 243;  return;
    }
    
    # n21221 frequency should be less than or equal to total5grams
    if ($n21221> $nppppp) {
	$errorMessage = "Frequency value ($n21221) must not exceed total number of 5grams.";
	$errorCodeNumber = 244;  return;
    }

    # n21222 should be greater than equal to zero 
    if ($n21222< 0)    {
	$errorMessage = "Frequency value ($n21222) must not be negative.";
	$errorCodeNumber = 245;  return;
    }
    
    # n21222 frequency should be less than or equal to total5grams
    if ($n21222> $nppppp) {
	$errorMessage = "Frequency value ($n21222) must not exceed total number of 5grams.";
	$errorCodeNumber = 246;  return;
    }

    # n22111 should be greater than equal to zero 
    if ($n22111< 0)    {
	$errorMessage = "Frequency value ($n22111) must not be negative.";
	$errorCodeNumber = 247;  return;
    }
    
    # n22111 frequency should be less than or equal to total5grams
    if ($n22111> $nppppp) {
	$errorMessage = "Frequency value ($n22111) must not exceed total number of 5grams.";
	$errorCodeNumber = 248;  return;
    }

    # n22112 should be greater than equal to zero 
    if ($n22112< 0)    {
	$errorMessage = "Frequency value ($n22112) must not be negative.";
	$errorCodeNumber = 249;  return;
    }
    
    # n22112 frequency should be less than or equal to total5grams
    if ($n22112> $nppppp) {
	$errorMessage = "Frequency value ($n22112) must not exceed total number of 5grams.";
	$errorCodeNumber = 250;  return;
    }

    # n22121 should be greater than equal to zero 
    if ($n22121< 0)    {
	$errorMessage = "Frequency value ($n22121) must not be negative.";
	$errorCodeNumber = 251;  return;
    }
    
    # n22121 frequency should be less than or equal to total5grams
    if ($n22121> $nppppp) {
	$errorMessage = "Frequency value ($n22121) must not exceed total number of 5grams.";
	$errorCodeNumber = 252;  return;
    }

    # n22122 should be greater than equal to zero 
    if ($n22122< 0)    {
	$errorMessage = "Frequency value ($n22122) must not be negative.";
	$errorCodeNumber = 253;  return;
    }
    
    # n22122 frequency should be less than or equal to total5grams
    if ($n22122> $nppppp) {
	$errorMessage = "Frequency value ($n22122) must not exceed total number of 5grams.";
	$errorCodeNumber = 254;  return;
    }
    
    # n22211 should be greater than equal to zero 
    if ($n22211< 0)    {
	$errorMessage = "Frequency value ($n22211) must not be negative.";
	$errorCodeNumber = 255;  return;
    }
    
    # n22211 frequency should be less than or equal to total5grams
    if ($n22211> $nppppp) {
	$errorMessage = "Frequency value ($n22211) must not exceed total number of 5grams.";
	$errorCodeNumber = 256;  return;
    }

    # n22212 should be greater than equal to zero 
    if ($n22212< 0)    {
	$errorMessage = "Frequency value ($n22212) must not be negative.";
	$errorCodeNumber = 257;  return;
    }
    
    # n22212 frequency should be less than or equal to total5grams
    if ($n22212> $nppppp) {
	$errorMessage = "Frequency value ($n22212) must not exceed total number of 5grams.";
	$errorCodeNumber = 258;  return;
    }
    
    # n22221 should be greater than equal to zero 
    if ($n22221< 0)    {
	$errorMessage = "Frequency value ($n22221) must not be negative.";
	$errorCodeNumber = 259;  return;
    }
    
    # n22221 frequency should be less than or equal to total5grams
    if ($n22221> $nppppp) {
	$errorMessage = "Frequency value ($n22221) must not exceed total number of 5grams.";
	$errorCodeNumber = 260;  return;
    }
    
    # n22222 should be greater than equal to zero 
    if ($n22222< 0)    {
	$errorMessage = "Frequency value ($n22222) must not be negative.";
	$errorCodeNumber = 261;  return;
    }
    
    # n22222 frequency should be less than or equal to total5grams
    if ($n22222> $nppppp) {
	$errorMessage = "Frequency value ($n22222) must not exceed total number of 5grams.";
	$errorCodeNumber = 262;  return;
    }
    
    # Everything looks good so we can return the observed values
    @values = ($n11111, $n11112, $n11121, $n11122, $n11211, $n11212, $n11221, 
	       $n11222, $n12111, $n12112, $n12121, $n12122, $n12211, $n12212, 
	       $n12221, $n12222, $n21111, $n21112, $n21121, $n21122, $n21211, 
	       $n21212, $n21221, $n21222, $n22111, $n22112, $n22121, $n22122, 
	       $n22211, $n22212, $n22221, $n22222); 

    return @values;
}

#  Calculates, checks and returns the expected values
sub getExpectedValues
{
    #  Calculate the default expected values
    $m11111=$n1pppp*$np1ppp*$npp1pp*$nppp1p*$npppp1/($nppppp**4);
    $m11112=$n1pppp*$np1ppp*$npp1pp*$nppp1p*$npppp2/($nppppp**4);
    $m11121=$n1pppp*$np1ppp*$npp1pp*$nppp2p*$npppp1/($nppppp**4);
    $m11122=$n1pppp*$np1ppp*$npp1pp*$nppp2p*$npppp2/($nppppp**4);
    $m11211=$n1pppp*$np1ppp*$npp2pp*$nppp1p*$npppp1/($nppppp**4);
    $m11212=$n1pppp*$np1ppp*$npp2pp*$nppp1p*$npppp2/($nppppp**4);
    $m11221=$n1pppp*$np1ppp*$npp2pp*$nppp2p*$npppp1/($nppppp**4);
    $m11222=$n1pppp*$np1ppp*$npp2pp*$nppp2p*$npppp2/($nppppp**4);
    $m12111=$n1pppp*$np2ppp*$npp1pp*$nppp1p*$npppp1/($nppppp**4);
    $m12112=$n1pppp*$np2ppp*$npp1pp*$nppp1p*$npppp2/($nppppp**4);
    $m12121=$n1pppp*$np2ppp*$npp1pp*$nppp2p*$npppp1/($nppppp**4);
    $m12122=$n1pppp*$np2ppp*$npp1pp*$nppp2p*$npppp2/($nppppp**4);
    $m12211=$n1pppp*$np2ppp*$npp2pp*$nppp1p*$npppp1/($nppppp**4);
    $m12212=$n1pppp*$np2ppp*$npp2pp*$nppp1p*$npppp2/($nppppp**4);
    $m12221=$n1pppp*$np2ppp*$npp2pp*$nppp2p*$npppp1/($nppppp**4);
    $m12222=$n1pppp*$np2ppp*$npp2pp*$nppp2p*$npppp2/($nppppp**4);
    $m21111=$n2pppp*$np1ppp*$npp1pp*$nppp1p*$npppp1/($nppppp**4);
    $m21112=$n2pppp*$np1ppp*$npp1pp*$nppp1p*$npppp2/($nppppp**4);
    $m21121=$n2pppp*$np1ppp*$npp1pp*$nppp2p*$npppp1/($nppppp**4);
    $m21122=$n2pppp*$np1ppp*$npp1pp*$nppp2p*$npppp2/($nppppp**4);
    $m21211=$n2pppp*$np1ppp*$npp2pp*$nppp1p*$npppp1/($nppppp**4);
    $m21212=$n2pppp*$np1ppp*$npp2pp*$nppp1p*$npppp2/($nppppp**4);
    $m21221=$n2pppp*$np1ppp*$npp2pp*$nppp2p*$npppp1/($nppppp**4);
    $m21222=$n2pppp*$np1ppp*$npp2pp*$nppp2p*$npppp2/($nppppp**4);
    $m22111=$n2pppp*$np2ppp*$npp1pp*$nppp1p*$npppp1/($nppppp**4);
    $m22112=$n2pppp*$np2ppp*$npp1pp*$nppp1p*$npppp2/($nppppp**4);
    $m22121=$n2pppp*$np2ppp*$npp1pp*$nppp2p*$npppp1/($nppppp**4);
    $m22122=$n2pppp*$np2ppp*$npp1pp*$nppp2p*$npppp2/($nppppp**4);
    $m22211=$n2pppp*$np2ppp*$npp2pp*$nppp1p*$npppp1/($nppppp**4);
    $m22212=$n2pppp*$np2ppp*$npp2pp*$nppp1p*$npppp2/($nppppp**4);
    $m22221=$n2pppp*$np2ppp*$npp2pp*$nppp2p*$npppp1/($nppppp**4);
    $m22222=$n2pppp*$np2ppp*$npp2pp*$nppp2p*$npppp2/($nppppp**4);
    
    # dont want ($n11111 / $m11111) to be 0 or less! flag error if so!
    if ( $n11111 )  
    { 
	if ($m11111 == 0) {
	    $errorMessage = "Expected value in cell (1,1,1,1,1) must not be zero";
	    $errorCodeNumber = 231;  return;
	}

	if (($n11111 / $m11111) < 0) {
	    $errorMessage = "About to take log of negative value for cell (1,1,1,1,1)";
	    $errorCodeNumber = 232;  return;
	}
    }
    
    if ( $n11112 )  
    { 
	if ($m11112 == 0) {
	    $errorMessage = "Expected value in cell (1,1,1,1,2) must not be zero";
	    $errorCodeNumber = 233;  return;
	}
	
	if (($n11112 / $m11112) < 0) {
	    $errorMessage = "About to take log of negative value for cell (1,1,1,1,2)";
	    $errorCodeNumber = 234;  return;
	}
    }
    
    if ( $n11121 )  
    { 
	if ($m11121 == 0) {
	    $errorMessage = "Expected value in cell (1,1,1,2,1) must not be zero";
	    $errorCodeNumber = 235;  return;
	}

	if (($n11121 / $m11121) < 0) {
	    $errorMessage = "About to take log of negative value for cell (1,1,1,2,1)";
	    $errorCodeNumber = 236;  return;
	}
    }

    if ( $n11122 )  
    { 
	if ($m11122 == 0) {
	    $errorMessage = "Expected value in cell (1,1,1,2,2) must not be zero";
	    $errorCodeNumber = 237;  return;
	}

	if (($n11122 / $m11122) < 0) {
	    $errorMessage = "About to take log of negative value for cell (1,1,1,2,2)";
	    $errorCodeNumber = 238;  return;
	}
    }
    
    if ( $n11211 )  
    { 
	if ($m11211 == 0) {
	    $errorMessage = "Expected value in cell (1,1,2,1,1) must not be zero";
	    $errorCodeNumber = 239;  return;
	}

	if (($n11211 / $m11211) < 0) {
	    $errorMessage = "About to take log of negative value for cell (1,1,2,1,1)";
	    $errorCodeNumber = 240;  return;
	}
    }

    if ( $n11212 )  
    { 
	if ($m11212 == 0) {
	    $errorMessage = "Expected value in cell (1,1,2,1,2) must not be zero";
	    $errorCodeNumber = 241;  return;
	}

	if (($n11212 / $m11212) < 0) {
	    $errorMessage = "About to take log of negative value for cell (1,1,2,1,2)";
	    $errorCodeNumber = 242;  return;
	}
    }

     if ( $n11221 )  
    { 
	if ($m11221 == 0) {
	    $errorMessage = "Expected value in cell (1,1,2,2,1) must not be zero";
	    $errorCodeNumber = 243;  return;
	}

	if (($n11221 / $m11221) < 0) {
	    $errorMessage = "About to take log of negative value for cell (1,1,2,2,1)";
	    $errorCodeNumber = 243;  return;
	}
    }

    if ( $n11222 )  
    { 
	if ($m11222 == 0) {
	    $errorMessage = "Expected value in cell (1,1,2,2,2) must not be zero";
	    $errorCodeNumber = 244;  return;
	}

	if (($n11222 / $m11222) < 0) {
	    $errorMessage = "About to take log of negative value for cell (1,1,2,2,2)";
	    $errorCodeNumber = 245;  return;
	}
    }

    if ( $n12111 )  
    { 
	if ($m12111 == 0) {
	    $errorMessage = "Expected value in cell (1,2,1,1,1) must not be zero";
	    $errorCodeNumber = 246;  return;
	}

	if (($n12111 / $m12111) < 0) {
	    $errorMessage = "About to take log of negative value for cell (1,2,1,1,1)";
	    $errorCodeNumber = 247;  return;
	}
    }
    
    if ( $n12112 )  
    { 
	if ($m12112 == 0) {
	    $errorMessage = "Expected value in cell (1,2,1,1,2) must not be zero";
	    $errorCodeNumber = 248;  return;
	}

	if (($n12112 / $m12112) < 0) {
	    $errorMessage = "About to take log of negative value for cell (1,2,1,1,2)";
	    $errorCodeNumber = 249;  return;
	}
    }

    if ( $n12121 )  
    { 
	if ($m12121 == 0) {
	    $errorMessage = "Expected value in cell (1,2,1,2,1) must not be zero";
	    $errorCodeNumber = 250;  return;
	}

	if (($n12121 / $m12121) < 0) {
	    $errorMessage = "About to take log of negative value for cell (1,2,1,2,1)";
	    $errorCodeNumber = 251;  return;
	}
    }

    if ( $n12122 )  
    { 
	if ($m12122 == 0) {
	    $errorMessage = "Expected value in cell (1,2,1,2,2) must not be zero";
	    $errorCodeNumber = 252;  return;
	}

	if (($n12122 / $m12122) < 0) {
	    $errorMessage = "About to take log of negative value for cell (1,2,1,2,2)";
	    $errorCodeNumber = 253;  return;
	}
    }

    if ( $n12211 )  
    { 
	if ($m12211 == 0) {
	    $errorMessage = "Expected value in cell (1,2,2,1,1) must not be zero";
	    $errorCodeNumber = 254;  return;
	}

	if (($n12211 / $m12211) < 0) {
	    $errorMessage = "About to take log of negative value for cell (1,2,2,1,1)";
	    $errorCodeNumber = 255;  return;
	}
    }
    
    if ( $n12212 )  
    { 
	if ($m12212 == 0) {
	    $errorMessage = "Expected value in cell (1,2,2,1,2) must not be zero";
	    $errorCodeNumber = 256;  return;
	}

	if (($n12212 / $m12212) < 0) {
	    $errorMessage = "About to take log of negative value for cell (1,2,2,1,2)";
	    $errorCodeNumber = 257;  return;
	}
    }

     if ( $n12221 )  
    { 
	if ($m12221 == 0) {
	    $errorMessage = "Expected value in cell (1,2,2,2,1) must not be zero";
	    $errorCodeNumber = 258;  return;
	}

	if (($n12221 / $m12221) < 0) {
	    $errorMessage = "About to take log of negative value for cell (1,2,2,2,1)";
	    $errorCodeNumber = 259;  return;
	}
    }
    
     if ( $n12222 )  
    { 
	if ($m12222 == 0) {
	    $errorMessage = "Expected value in cell (1,2,2,2,2) must not be zero";
	    $errorCodeNumber = 260;  return;
	}

	if (($n12222 / $m12222) < 0) {
	    $errorMessage = "About to take log of negative value for cell (1,2,2,2,2)";
	    $errorCodeNumber = 261;  return;
	}
    }

    if ( $n21111 )  
    { 
	if ($m21111 == 0) {
	    $errorMessage = "Expected value in cell (2,1,1,1,1) must not be zero";
	    $errorCodeNumber = 262;  return;
	}

	if (($n21111 / $m21111) < 0) {
	    $errorMessage = "About to take log of negative value for cell (2,1,1,1,1)";
	    $errorCodeNumber = 263;  return;
	}
    }

    if ( $n21112 )  
    { 
	if ($m21112 == 0) {
	    $errorMessage = "Expected value in cell (2,1,1,1,2) must not be zero";
	    $errorCodeNumber = 264;  return;
	}

	if (($n21112 / $m21112) < 0) {
	    $errorMessage = "About to take log of negative value for cell (2,1,1,1,2)";
	    $errorCodeNumber = 265;  return;
	}
    }

    if ( $n21121 )  
    { 
	if ($m21121 == 0) {
	    $errorMessage = "Expected value in cell (2,1,1,2,1) must not be zero";
	    $errorCodeNumber = 266;  return;
	}

	if (($n21121 / $m21121) < 0) {
	    $errorMessage = "About to take log of negative value for cell (2,1,1,2,1)";
	    $errorCodeNumber = 267;  return;
	}
    }
    
    if ( $n21122 )  
    { 
	if ($m21122 == 0) {
	    $errorMessage = "Expected value in cell (2,1,1,2,2) must not be zero";
	    $errorCodeNumber = 268;  return;
	}

	if (($n21122 / $m21122) < 0) {
	    $errorMessage = "About to take log of negative value for cell (2,1,1,2,2)";
	    $errorCodeNumber = 269;  return;
	}
    }

    if ( $n21211 )  
    { 
	if ($m21211 == 0) {
	    $errorMessage = "Expected value in cell (2,1,2,1,1) must not be zero";
	    $errorCodeNumber = 270;  return;
	}

	if (($n21211 / $m21211) < 0) {
	    $errorMessage = "About to take log of negative value for cell (2,1,2,1,1)";
	    $errorCodeNumber = 271;  return;
	}
    }

    if ( $n21212 )  
    { 
	if ($m21212 == 0) {
	    $errorMessage = "Expected value in cell (2,1,2,1,2) must not be zero";
	    $errorCodeNumber = 272;  return;
	}

	if (($n21212 / $m21212) < 0) {
	    $errorMessage = "About to take log of negative value for cell (2,1,2,1,2)";
	    $errorCodeNumber = 273;  return;
	}
    }

    if ( $n21221 )  
    { 
	if ($m21221 == 0) {
	    $errorMessage = "Expected value in cell (2,1,2,2,1) must not be zero";
	    $errorCodeNumber = 274;  return;
	}

	if (($n21221 / $m21221) < 0) {
	    $errorMessage = "About to take log of negative value for cell (2,1,2,2,1)";
	    $errorCodeNumber = 275;  return;
	}
    }

    if ( $n21222 )  
    { 
	if ($m21222 == 0) {
	    $errorMessage = "Expected value in cell (2,1,2,2,2) must not be zero";
	    $errorCodeNumber = 276;  return;
	}

	if (($n21222 / $m21222) < 0) {
	    $errorMessage = "About to take log of negative value for cell (2,1,2,2,2)";
	    $errorCodeNumber = 277;  return;
	}
    }

    if ( $n22111 )  
    { 
	if ($m22111 == 0) {
	    $errorMessage = "Expected value in cell (2,2,1,1,1) must not be zero";
	    $errorCodeNumber = 278;  return;
	}

	if (($n22111 / $m22111) < 0) {
	    $errorMessage = "About to take log of negative value for cell (2,2,1,1,1)";
	    $errorCodeNumber = 279;  return;
	}
    }

    if ( $n22112 )  
    { 
	if ($m22112 == 0) {
	    $errorMessage = "Expected value in cell (2,2,1,1,2) must not be zero";
	    $errorCodeNumber = 280;  return;
	}

	if (($n22112 / $m22112) < 0) {
	    $errorMessage = "About to take log of negative value for cell (2,2,1,1,2)";
	    $errorCodeNumber = 281;  return;
	}
    }

    if ( $n22121 )  
    { 
	if ($m22121 == 0) {
	    $errorMessage = "Expected value in cell (2,2,1,2,1) must not be zero";
	    $errorCodeNumber = 282;  return;
	}

	if (($n22121 / $m22121) < 0) {
	    $errorMessage = "About to take log of negative value for cell (2,2,1,2,1)";
	    $errorCodeNumber = 283;  return;
	}
    }

    if ( $n22122 )  
    { 
	if ($m22122 == 0) {
	    $errorMessage = "Expected value in cell (2,2,1,2,2) must not be zero";
	    $errorCodeNumber = 284;  return;
	}

	if (($n22122 / $m22122) < 0) {
	    $errorMessage = "About to take log of negative value for cell (2,2,1,2,2)";
	    $errorCodeNumber = 285;  return;
	}
    }

    if ( $n22211 )  
    { 
	if ($m22211 == 0) {
	    $errorMessage = "Expected value in cell (2,2,2,1,1) must not be zero";
	    $errorCodeNumber = 286;  return;
	}

	if (($n22211 / $m22211) < 0) {
	    $errorMessage = "About to take log of negative value for cell (2,2,2,1,1)";
	    $errorCodeNumber = 287;  return;
	}
    }

    if ( $n22212 )  
    { 
	if ($m22212 == 0) {
	    $errorMessage = "Expected value in cell (2,2,2,1,2) must not be zero";
	    $errorCodeNumber = 288;  return;
	}

	if (($n22212 / $m22212) < 0) {
	    $errorMessage = "About to take log of negative value for cell (2,2,2,1,2)";
	    $errorCodeNumber = 289;  return;
	}
    }

    if ( $n22221 )  
    { 
	if ($m22221 == 0) {
	    $errorMessage = "Expected value in cell (2,2,2,2,1) must not be zero";
	    $errorCodeNumber = 290;  return;
	}

	if (($n22221 / $m22221) < 0) {
	    $errorMessage = "About to take log of negative value for cell (2,2,2,2,1)";
	    $errorCodeNumber = 291;  return;
	}
    }

    if ( $n22222 )  
    { 
	if ($m22222 == 0) {
	    $errorMessage = "Expected value in cell (2,2,2,2,2) must not be zero";
	    $errorCodeNumber = 292;  return;
	}

	if (($n22222 / $m22222) < 0) {
	    $errorMessage = "About to take log of negative value for cell (2,2,2,2,2)";
	    $errorCodeNumber = 293;  return;
	}
    }

    #  Everything looks good so we can return the expected values
    @values = ($m11111, $m11112, $m11121, $m11122, $m11211, $m11212, $m11221, 
	       $m11222, $m12111, $m12112, $m12121, $m12122, $m12211, $m12212, 
	       $m12221, $m12222, $m21111, $m21112, $m21121, $m21122, $m21211, 
	       $m21212, $m21221, $m21222, $m22111, $m22112, $m22121, $m22122, 
	       $m22211, $m22212, $m22221, $m22222);
    
    return @values;
}

#  Returns the marginal frequencies
sub getMarginalTotals {
    @values = ( $npppp1, $nppp1p, $nppp11, $npp1pp, $npp1p1, $npp11p, 
		$npp111, $np1ppp, $np1pp1, $np1p1p, $np1p11, $np11pp, 
		$np11p1, $np111p, $np1111, $n1pppp, $n1ppp1, $n1pp1p, 
		$n1pp11, $n1p1pp, $n1p1p1, $n1p11p, $n1p111, $n11ppp, 
		$n11pp1, $n11p1p, $n11p11, $n111pp, $n111p1, $n1111p );
    
    return @values;
}

#  Return the total number of 5grams
sub getTotal5grams {
    return $nppppp;
}

# function to return the error code of the last operation and reset
# error code. useful if the error can be recovered from!
sub errorCode 
{ 
    my $temp = $errorCodeNumber;
    $errorCodeNumber = 0;
    return($temp); 
}


# function to return the error message of the last operation and reset
# the message string. useful if error can be recovered from!
sub errorString
{
    my $temp = $errorMessage;
    $errorMessage = "";
    return($temp);
}

1;
